"""Configuration module."""

# gCO2eq/kWh - source: https://ourworldindata.org/grapher/carbon-intensity-electricity Global Average
# Currency/kWh (Euro) - source: https://www.stromauskunft.de/strompreise/ 03.05.2023
import configparser
import logging
import os
import pprint as pp
from pathlib import Path
from typing import Any, Mapping

from perun.io.io import IOFormat

log = logging.getLogger(__name__)

_default_config: Mapping[str, Mapping[str, Any]] = {
    "post-processing": {
        "power_overhead": 0,  # Watt
        "pue": 1.0,  # No assumption where the workflow is running
        "emissions_factor": 230.0,  # gCO2eq/kWh - Germany, 20.05.2025 (https://app.electricitymaps.com/zone/DE/72h/hourly)
        "price_factor": 0.2678,  # Euro/kWh - Germany, 20.05.2025 (https://www.stromauskunft.de/strompreise/)
        "price_unit": "€",
    },
    "monitor": {
        "sampling_period": 1,
        "include_backends": "",
        "include_sensors": "",
        "exclude_backends": "",
        "exclude_sensors": "",
    },
    "output": {
        "app_name": None,
        "run_id": None,
        "format": "text",
        "data_out": "./perun_results",
    },
    "benchmarking": {
        "rounds": 1,
        "warmup_rounds": 0,
        "metrics": "runtime,energy",
        "region_metrics": "runtime,power",
    },
    "benchmarking.units": {
        "joule": "k",
        "second": "",
        "percent": "",
        "watt": "",
        "byte": "G",
    },
    "debug": {"log_lvl": "WARNING", "log_file": None},
}

config: configparser.ConfigParser = configparser.ConfigParser(allow_no_value=True)
config.read_dict(_default_config)

globalConfigPath = Path.home() / ".config/perun.ini"
if globalConfigPath.exists() and globalConfigPath.is_file():
    config.read(globalConfigPath)

localConfigPath = Path.cwd() / ".perun.ini"
if globalConfigPath.exists() and globalConfigPath.is_file():
    config.read(globalConfigPath)


def read_custom_config(
    pathStr: str,
) -> None:
    """Read INI config file and save in global configuration objects.

    Parameters
    ----------
    pathStr : str
        Path to configuration file.
    """
    configPath: Path = Path(pathStr)
    if configPath.exists() and configPath.is_file():
        config.read(configPath)


def save_to_config(key: str, value: Any) -> None:
    """Save key and value to configuration.

    Parameters
    ----------
    key : str
        Option name
    value : Any
        Option value
    """
    for section in config.sections():
        if config.has_option(section, key):
            config.set(section, key, str(value))


def read_environ() -> None:
    """Read perun environmental variables."""
    for section, subconf in _default_config.items():
        for option in subconf.keys():
            envvar = f"PERUN_{option.upper()}"
            if envvar in os.environ:
                config.set(section, option, os.environ[envvar])


def sanitize_config(config: configparser.ConfigParser) -> configparser.ConfigParser:
    """Sanitize configuration values.

    Parameters
    ----------
    config : configparser.ConfigParser
        Configuration object.

    Returns
    -------
    configparser.ConfigParser
        Sanitized configuration object.
    """
    # Ensure post processing variables are valid
    try:
        power_overhead = config.getfloat("post-processing", "power_overhead")
        if power_overhead < 0:
            log.warning(
                f"Invalid power overhead {power_overhead}. Should be a number higher or equal than 0. Defaulting to 0."
            )
            config.set("post-processing", "power_overhead", "0")
    except ValueError:
        log.warning(
            "Invalid power overhead. Should be a number higher or equal than 0. Defaulting to 0."
        )
        config.set("post-processing", "power_overhead", "0")

    try:
        pue = config.getfloat("post-processing", "pue")
        if pue < 1:
            log.warning(
                f"Invalid PUE {pue}. Should be a number higher or equal than 1. Defaulting to 1."
            )
            config.set("post-processing", "pue", "1.0")
    except ValueError:
        log.warning(
            "Invalid PUE. Should be a number higher or equal than 1. Defaulting to 1."
        )
        config.set("post-processing", "pue", "1.0")

    try:
        default_emissions_factor = _default_config["post-processing"][
            "emissions_factor"
        ]
        emissions_factor = config.getfloat("post-processing", "emissions_factor")
        if emissions_factor < 0:
            # Default value is the global average emissions factor
            log.warning(
                f"Invalid emissions factor {emissions_factor}. Should be a number higher or equal than 0. Defaulting to {default_emissions_factor} gCO2eq/kWh."
            )
            config.set(
                "post-processing", "emissions_factor", str(default_emissions_factor)
            )
    except ValueError:
        log.warning(
            f"Invalid emissions factor. Should be a number higher or equal than 0. Defaulting to {default_emissions_factor} gCO2eq/kWh."
        )
        config.set("post-processing", "emissions_factor", str(default_emissions_factor))

    try:
        default_price_factor = _default_config["post-processing"]["price_factor"]
        price_factor = config.getfloat("post-processing", "price_factor")
        if price_factor < 0:
            log.warning(
                f"Invalid price factor {price_factor}. Should be a number higher or equal than 0. Defaulting to {default_price_factor} Currency/kWh."
            )
            config.set("post-processing", "price_factor", str(default_price_factor))
    except ValueError:
        log.warning(
            "Invalid price factor. Should be a number higher or equal than 0. Defaulting to 0.3251 Currency/kWh."
        )
        config.set(
            "post-processing",
            "price_factor",
            str(_default_config["post-processing"]["price_factor"]),
        )
    # Ensure that the monitoring options are valid
    try:
        sampling_period = config.getfloat("monitor", "sampling_period")
        if sampling_period < 0.1:
            log.warning(
                f"Invalid sampling period {sampling_period}. Should be a number higher than 0.1 . Defaulting to 1."
            )
            config.set("monitor", "sampling_period", "1")
    except ValueError:
        log.warning(
            "Invalid sampling period. Should be a number higher than 0.1 . Defaulting to 1."
        )
        config.set("monitor", "sampling_period", "1")

    # Ensure only the include or exclude options are set
    include_backends = config.get("monitor", "include_backends")
    include_sensors = config.get("monitor", "include_sensors")
    exclude_backends = config.get("monitor", "exclude_backends")
    exclude_sensors = config.get("monitor", "exclude_sensors")

    if include_backends and exclude_backends:
        log.warning(
            "Both include and exclude backends options are set. Defaulting to exclude only."
        )
        config.set("monitor", "include_backends", "")

    if include_sensors and exclude_sensors:
        log.warning(
            "Both include and exclude sensors options are set. Defaulting to exclude only."
        )
        config.set("monitor", "include_sensors", "")

    # Ensure that the output format is valid
    try:
        format = config.get("output", "format")
        IOFormat(format)
    except ValueError:
        log.warning(
            f"Invalid output format {format}. Defaulting to text. Avilable formats: {pp.pformat([f.value for f in IOFormat])}"
        )
        config.set("output", "format", IOFormat.TEXT.value)

    # Ensure that the rounds and warmup rounds are valid
    try:
        rounds = config.getint("benchmarking", "rounds")
        if rounds < 1:
            log.warning(
                f"Invalid number rounds {rounds}. Should be a number higher than 1. Defaulting to 1."
            )
            config.set("benchmarking", "rounds", "1")
    except ValueError:
        log.warning(
            "Invalid number rounds. Should be a number higher than 1. Defaulting to 1."
        )
        config.set("benchmarking", "rounds", "1")

    try:
        warmup_rounds = config.getint("benchmarking", "warmup_rounds")
        if warmup_rounds < 0:
            log.warning(
                f"Invalid number warmup rounds {warmup_rounds}. Should be a number higher than 0. Defaulting to 0."
            )
            config.set("benchmarking", "warmup_rounds", "0")
    except ValueError:
        log.warning(
            "Invalid number warmup rounds. Should be a number higher than 0. Defaulting to 0."
        )
        config.set("benchmarking", "warmup_rounds", "0")

    # Ensure that the log_lvl is valid
    log_lvl = config.get("debug", "log_lvl")
    if log_lvl not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        log.warning(f"Invalid log level {log_lvl}. Defaulting to WARNING.")
        config.set("debug", "log_lvl", "WARNING")

    return config



import perun
import os

class perun2mlflow():

    def __init__(self):
        self._activated = False

    def __call__(self):
        self._activated = True
        import platform
        import mlflow
        import os
        hostname = platform.node()

        print(f"In callback: {hostname}")
        mlflow_uri = os.getenv("MLFLOW_TRACKING_URI", "http://localhost:5000")
        mlflow.set_tracking_uri(mlflow_uri)
        mlflow.set_experiment("perun-live-monitoring")

        mlflow.start_run(log_system_metrics=True)

        def func(values):
            """
            Function to log metrics to MLflow.
            """
            mlflow.log_metrics(values, synchronous=False)
        return func

    def __del__(self):
        if self._activated:
            import mlflow
            print("Closing run!")
            mlflow.end_run()

perun.register_live_callback(perun2mlflow(), "mlflow_callback")



import configparser
import json
import subprocess
from pathlib import Path

import pytest

from perun.api.cli import _get_arg_parser
from perun.core import Perun
from perun.io.text_report import sensors_table


def test_no_subcommand():
    processOut = subprocess.run(["perun"], capture_output=True, text=True, timeout=10)
    expectedResult = _get_arg_parser().format_help()
    assert processOut.stdout == expectedResult


@pytest.mark.parametrize(
    "flag, by_rank",
    [
        ([""], False),
        (["--all"], False),
        (["--by_rank"], True),
        (["--active"], True),
    ],
)
def test_sensors_command(flag, by_rank, perun: Perun):
    processOut = subprocess.run(
        ["perun", "sensors"] + flag, capture_output=True, text=True, timeout=10
    ).stdout.rstrip()
    expectedResult = sensors_table(
        (
            perun.g_available_sensors
            if flag[0] != "--active"
            else perun.g_assigned_sensors
        ),
        by_rank=by_rank,
    ).rstrip()
    assert processOut == expectedResult


def test_showconf_command(defaultConfig: configparser.ConfigParser):
    # 1) Are the outputs the same?
    processorOut = subprocess.run(
        ["perun", "showconf"], capture_output=True, text=True
    ).stdout
    parser = configparser.ConfigParser(allow_no_value=True)
    parser.read_string(processorOut)
    assert parser == defaultConfig


def test_showconf_command_with_cli_args(defaultConfig: configparser.ConfigParser):
    # 2) Are cli arguments correctly set?
    processorOut = subprocess.run(
        ["perun", "--log_lvl", "ERROR", "showconf"], capture_output=True, text=True
    ).stdout
    parser = configparser.ConfigParser(allow_no_value=True)
    parser.read_string(processorOut)
    defaultConfig.set("debug", "log_lvl", "ERROR")
    assert parser == defaultConfig


def test_showconf_command_with_conf_file(
    defaultConfig: configparser.ConfigParser, tmp_path: Path
):
    # 3) Are files read correctly?
    confPath = tmp_path / ".perun.ini"
    defaultConfig.set("monitor", "sampling_period", "2")
    with open(confPath, "w+") as configFile:
        defaultConfig.write(configFile)

    processorOut = subprocess.run(
        ["perun", "--configuration", str(confPath), "showconf"],
        capture_output=True,
        text=True,
    ).stdout
    parser = configparser.ConfigParser(allow_no_value=True)
    parser.read_string(processorOut)
    assert parser.get("monitor", "sampling_period") == "2"
    assert parser == defaultConfig


def test_showconf_command_with_default(
    defaultConfig: configparser.ConfigParser, tmp_path: Path
):
    # 4) Does default ignore everything?
    confPath = tmp_path / ".perun.ini"
    defaultConfig.set("monitor", "sampling_period", "2")
    with open(confPath, "w+") as configFile:
        defaultConfig.write(configFile)

    processorOut = subprocess.run(
        ["perun", "--log_lvl", "ERROR", "--configuration", str(confPath), "showconf"],
        capture_output=True,
        text=True,
    ).stdout
    print(processorOut)
    parser = configparser.ConfigParser(allow_no_value=True)
    parser.read_string(processorOut)
    assert defaultConfig.get("monitor", "sampling_period") == "2"
    assert defaultConfig.get("debug", "log_lvl") == "WARNING"
    assert parser.get("monitor", "sampling_period") == "2"
    assert parser.get("debug", "log_lvl") == "ERROR"
    assert parser != defaultConfig

    defaultConfig.set("monitor", "sampling_period", "1")
    processorOut = subprocess.run(
        [
            "perun",
            "--log_lvl",
            "ERROR",
            "--configuration",
            str(confPath),
            "showconf",
            "--default",
        ],
        capture_output=True,
        text=True,
    ).stdout
    print(processorOut)
    parser = configparser.ConfigParser(allow_no_value=True)
    parser.read_string(processorOut)

    assert defaultConfig.get("monitor", "sampling_period") == "1"
    assert defaultConfig.get("debug", "log_lvl") == "WARNING"
    assert parser.get("debug", "log_lvl") == "WARNING"
    assert parser.get("monitor", "sampling_period") == "1"
    assert parser == defaultConfig


def test_metadata_command(perun: Perun):
    processorOut = subprocess.run(
        ["perun", "metadata"], capture_output=True, text=True, timeout=10
    ).stdout
    metadataJson = json.loads(processorOut)
    for host in perun.host_rank.keys():
        assert host in metadataJson


def test_monitor_command(tmp_path: Path):
    # Test Monitor
    testFilePath = tmp_path / "idle.py"
    with open(testFilePath, "w+") as testFile:
        testFile.write("import time\n\ntime.sleep(10)")

    resultsPath = tmp_path / "results"

    subprocess.run(
        f"perun monitor --data_out {resultsPath} {testFilePath}".split(" "), timeout=20
    )

    # Expected files, hdf5 file and a text file with a date
    # Are the files in the correct folder
    resultFiles = list(resultsPath.iterdir())
    assert len(resultFiles) == 2
    assert resultsPath / "idle.hdf5" in resultFiles
    assert (resultsPath / "idle.hdf5").is_file()

    resultFiles.remove(resultsPath / "idle.hdf5")
    textFile = resultFiles.pop()
    assert textFile.is_file()
    assert textFile.suffix == ".txt"


def test_monitor_binary_command(tmp_path: Path):
    # Test Monitor
    resultsPath = tmp_path / "results"

    subprocess.run(
        f"perun monitor --data_out {resultsPath} --binary sleep 10".split(" "),
        timeout=20,
    )

    # Expected files, hdf5 file and a text file with a date
    # Are the files in the correct folder
    resultFiles = list(resultsPath.iterdir())
    assert len(resultFiles) == 2
    assert resultsPath / "sleep.hdf5" in resultFiles
    assert (resultsPath / "sleep.hdf5").is_file()

    resultFiles.remove(resultsPath / "sleep.hdf5")
    textFile = resultFiles.pop()
    assert textFile.is_file()
    assert textFile.suffix == ".txt"


@pytest.mark.parametrize(
    "format,suffix", [("json", "json"), ("csv", "csv"), ("pickle", "pkl")]
)
def test_export_command(format: str, suffix: str, tmp_path: Path):
    testFilePath = tmp_path / "idle.py"
    with open(testFilePath, "w+") as testFile:
        testFile.write("import time\n\ntime.sleep(10)")

    resultsPath = tmp_path / "results"
    subprocess.run(
        f"perun monitor --data_out {resultsPath} {testFilePath}".split(" "), timeout=20
    )

    # Expected files, hdf5 file and a text file, and a csv file
    # Are the files in the correct folder
    resultFiles = list(resultsPath.iterdir())
    assert len(resultFiles) == 2
    assert resultsPath / "idle.hdf5" in resultFiles

    resultFiles.remove(resultsPath / "idle.hdf5")
    textFile = resultFiles.pop()
    assert textFile.is_file()
    assert textFile.suffix == ".txt"

    # Test export
    subprocess.run(
        f"perun export {resultsPath / 'idle.hdf5'} {format}".split(" "), timeout=10
    )
    resultFiles = list(resultsPath.iterdir())
    assert len(resultFiles) == 3
    assert resultsPath / "idle.hdf5" in resultFiles
    assert textFile in resultFiles

    resultFiles.remove(resultsPath / "idle.hdf5")
    resultFiles.remove(textFile)

    exportedFile = resultFiles.pop()
    assert exportedFile.is_file()
    assert exportedFile.suffix == f".{suffix}"



"""Command line API."""

import argparse
import json
import logging
import sys
from pathlib import Path

import perun
from perun.configuration import (
    config,
    read_custom_config,
    read_environ,
    sanitize_config,
    save_to_config,
)
from perun.core import Perun
from perun.io.io import IOFormat
from perun.io.text_report import sensors_table
from perun.logging import set_logger_config
from perun.monitoring.application import Application

log = logging.getLogger(__name__)


def _get_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="perun",
        description="Distributed performance and energy monitoring tool",
        allow_abbrev=False,
    )
    parser.add_argument(
        "-c",
        "--configuration",
        default="./.perun.ini",
        help="Path to perun configuration file.",
    )
    parser.add_argument(
        "-l",
        "--log_lvl",
        choices=["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        help="Logging level.",
    )
    parser.add_argument(
        "--log_file",
        default=None,
        help="Path to the log file. None by default. Writting to a file disables logging in stdout.",
    )
    parser.add_argument(
        "--version", action="version", version=f"perun {perun.__version__}"
    )
    subparsers = parser.add_subparsers(dest="subcommand")

    # showconf
    showconf_parser = subparsers.add_parser(
        "showconf", help="Print perun configuration in INI format."
    )
    showconf_parser.add_argument(
        "--default",
        action="store_true",
        help="Print the default configuration",
        dest="showconf_default",
    )
    showconf_parser.set_defaults(func=showconf)

    # sensors
    sensor_parser = subparsers.add_parser(
        "sensors", help="Print available sensors by host and rank."
    )
    sensor_group = sensor_parser.add_mutually_exclusive_group()
    sensor_group.add_argument(
        "--all", help="Print all available sensors.", action="store_true"
    )
    sensor_group.add_argument(
        "--by_rank",
        help="Print sensors by available on each rank.",
        action="store_true",
    )
    sensor_group.add_argument(
        "--active",
        help="Print active sensors by rank based on the configuration file.",
        action="store_true",
    )
    sensor_parser.set_defaults(func=sensors)

    # metadata
    metadata_parser = subparsers.add_parser(
        "metadata", help="Print available metadata."
    )
    metadata_parser.set_defaults(func=metadata)

    # export
    export_parser = subparsers.add_parser(
        "export", help="Export existing output file to another format."
    )
    export_parser.add_argument(
        "-i", "--id", dest="run_id", help="Run id to export, last one by default"
    )
    export_parser.add_argument(
        "input_file",
        help="Existing perun output file. Should be hdf5, json or pickle.",
    )
    export_parser.add_argument(
        "format",
        help="Desired data output format.",
    )
    export_parser.set_defaults(func=export)

    # monitor
    monitor_parser = subparsers.add_parser(
        "monitor",
        help="""
    Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is running.

    SCRIPT is a path to the python script to monitor, run with arguments SCRIPT_ARGS.
    """,
    )
    monitor_parser.add_argument(
        "-n",
        "--app_name",
        dest="app_name",
        help="Name o the monitored application. The name is used to distinguish between multiple application in the same directory. If left empty, the file name will be used.",
    )
    monitor_parser.add_argument(
        "-i",
        "--run_id",
        dest="run_id",
        help="Unique id of the latest run of the application. If left empty, perun will use the current date.",
    )
    monitor_parser.add_argument(
        "-f",
        "--format",
        help="Secondary report format.",
        choices=[format.value for format in IOFormat],
    )
    monitor_parser.add_argument(
        "--data_out",
        help="Directory where output files are saved. Defaults to ./perun_results",
    )
    monitor_parser.add_argument(
        "--sampling_period",
        type=float,
        help="Sampling period in seconds. Defaults to 1 second.",
    )
    monitor_parser.add_argument(
        "--include_sensors",
        type=str,
        help="Comma separated list of sensors to use. Defaults to an empty string (all available sensors). ",
    )
    monitor_parser.add_argument(
        "--include_backends",
        type=str,
        help="Comma separated list of measuring backends to use. Defaults to an empty string (all available sensors).",
    )
    monitor_parser.add_argument(
        "--exclude_sensors",
        type=str,
        help="Comma separated list of sensors to exclude. Defaults to an empty string (all available sensors).",
    )
    monitor_parser.add_argument(
        "--exclude_backends",
        type=str,
        help="Comma separated list of measuring backends to use. Defaults to an empty string (all available sensors).",
    )
    monitor_parser.add_argument(
        "--power_overhead",
        type=float,
        help="Estimated power consumption of non-measured hardware components in Watts. Will be added to measured power consumption on the text report summary. Defaults to 0 Watts",
    )
    monitor_parser.add_argument(
        "--pue", type=float, help="Data center Power Usage Effectiveness. Defaults to 1"
    )
    monitor_parser.add_argument(
        "--price_factor",
        type=float,
        help="Electricity to Currency convertion factor in the form of Currency/kWh. Defaults to 0.3251 €/kWh",
    )
    monitor_parser.add_argument(
        "--price_unit",
        type=str,
        help="Currency character to use on the text report summary. Defaults to €",
    )
    monitor_parser.add_argument(
        "--emission_factor",
        type=float,
        help="Average carbon intensity of electricity (gCO2e/kWh). Defaults to 417.80 gC02e/kWh",
    )
    monitor_parser.add_argument(
        "--rounds", type=int, help="Number of warmup rounds to run app. Defaults to 1"
    )
    monitor_parser.add_argument(
        "--warmup_rounds",
        type=int,
        help="Number of warmup rounds to run the app. A warmup round is a full run of the application without gathering performance data. Defaults to 0",
    )
    monitor_parser.add_argument(
        "--bench_metrics",
        dest="metrics",
        type=str,
        help="List of metrics to add to the benchmark results. Only relevant when using the 'bench' format. Defaults to 'runtime,energy'",
    )
    monitor_parser.add_argument(
        "--region_metrics",
        type=str,
        help="List of metrics to add to the benchmark results that are associated with individual regions. Only relevant when using the 'bench' format. Defaults to 'runtime,energy'",
    )
    monitor_parser.add_argument(
        "-b",
        "--binary",
        action="store_true",
        help="Indicate if the monitored application is a binary. Otherwise treat it as a python script.",
    )
    monitor_parser.add_argument(
        "--live_callback_files",
        nargs="+",
        default=[],
        dest="live_callback_files",
        type=argparse.FileType("r", encoding="utf-8"),
        help="Path to file that contains the live callbacks to use.",
    )
    monitor_parser.add_argument("cmd", type=str)
    monitor_parser.add_argument("cmd_args", nargs=argparse.REMAINDER)
    monitor_parser.set_defaults(func=monitor)
    return parser


def cli() -> None:
    """Command line entrypoint."""
    parser = _get_arg_parser()

    # parse and read conf file and env
    args, remaining = parser.parse_known_args()

    if args.subcommand is None:
        parser.print_help()
        return

    # 1) Read custom configuration
    if args.configuration:
        read_custom_config(args.configuration)

    # 2) Read environment variables
    read_environ()

    # 3) Parse remaining arguments
    for key, value in vars(args).items():
        if value:
            save_to_config(key, value)

    sanitize_config(config)
    set_logger_config(config)

    # start function
    if hasattr(args, "func"):
        args.func(args)


def showconf(args: argparse.Namespace) -> None:
    """Print current perun configuration in INI format."""
    from perun.configuration import _default_config

    if args.showconf_default:
        config.read_dict(_default_config)
        config.write(sys.stdout)
    else:
        config.write(sys.stdout)


def sensors(args: argparse.Namespace) -> None:
    """Print available sensors."""
    perun = Perun(config)
    log.debug("Initialized perun object.")
    arg_by_rank = args.by_rank
    arg_active = args.active

    if arg_by_rank:
        log.debug("Printing sensors by rank.")
        g_available_sensors = perun.g_available_sensors
        if perun.comm.Get_rank() == 0:
            print(sensors_table(g_available_sensors))
    elif arg_active:
        log.debug("Printing active sensors by rank.")
        g_assigned_sensors = perun.g_assigned_sensors
        if perun.comm.Get_rank() == 0:
            print(sensors_table(g_assigned_sensors))
    else:
        log.debug("Printing all available sensors.")
        g_available_sensors = perun.g_available_sensors
        available_sensors: dict[str, tuple] = {}
        for _, sensors in enumerate(g_available_sensors):
            available_sensors.update(sensors)
        if perun.comm.Get_rank() == 0:
            print(sensors_table([available_sensors], by_rank=False))


def metadata(args: argparse.Namespace) -> None:
    """Print global metadata dictionaries in json format."""
    perun = Perun(config)

    hostMD = perun.l_host_metadata
    hostMD["backends"] = perun.l_backend_metadata
    allHostsMD = perun.comm.gather(hostMD, root=0)

    if perun.comm.Get_rank() == 0 and allHostsMD:
        metadataDict = {}
        for host, assignedRanks in perun.host_rank.items():
            metadataDict[host] = allHostsMD[assignedRanks[0]]

        json.dump(metadataDict, sys.stdout, indent=4)


def export(args: argparse.Namespace) -> None:
    """Export existing perun output file to another format."""
    in_file = Path(args.input_file)
    if not in_file.exists():
        log.error("File does not exist.")
        return

    perun = Perun(config)

    out_path = in_file.parent
    inputFormat = IOFormat.fromSuffix(in_file.suffix)
    out_format = IOFormat(args.format)

    dataNode = perun.import_from(in_file, inputFormat)
    if args.run_id:
        perun.export_to(out_path, dataNode, out_format, args.run_id)
    else:
        perun.export_to(out_path, dataNode, out_format)


def monitor(args: argparse.Namespace) -> None:
    """
    Gather power consumption from hardware devices while SCRIPT [SCRIPT_ARGS] is running.

    SCRIPT is a path to the python script to monitor, run with arguments SCRIPT_ARGS.
    """
    cmd: str = args.cmd
    log.debug(f"Cmd: {cmd}")
    argIndex = sys.argv.index(args.cmd)
    sys.argv = sys.argv[argIndex:]
    cmd_args: list[str] = sys.argv.copy()
    log.debug(f"Cmd args: {cmd_args}")
    if not args.binary:
        scriptPath = Path(cmd)
        try:
            assert scriptPath.exists()
            assert scriptPath.is_file()
            assert scriptPath.suffix == ".py"
        except AssertionError:
            log.error(
                f"Invalid script path. File {scriptPath} does not exist or is not a python script."
            )

        sys.path.insert(0, str(scriptPath.parent.absolute()))
        app = Application(scriptPath, config, args=tuple(sys.argv[1:]))
    else:
        app = Application(cmd, config, is_binary=True, args=tuple(sys.argv[1:]))

    perun = Perun(config)
    for live_callback_file in args.live_callback_files:
        log.info(f"Loading live callback from file: {live_callback_file.name}")
        try:
            # Load the files using importlib
            live_callback_code = live_callback_file.read()
            exec(live_callback_code, globals(), locals())
            log.info("Live callback loaded successfully.")
        except Exception as e:
            log.error(
                f"Error loading live callback from file {live_callback_file.name}: {e}"
            )
            continue

    log.info("Starting perun monitoring application.")

    perun.monitor_application(app)



import argparse

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from perun import monitor


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


@monitor()
def train(args, model, device, train_loader, test_loader, optimizer, scheduler):
    for epoch in range(1, args.epochs + 1):
        train_epoch(args, model, device, train_loader, optimizer, epoch)
        test(model, device, test_loader)
        scheduler.step()


@monitor()
def train_epoch(args, model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )
            if args.dry_run:
                break


@monitor()
def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(
                output, target, reduction="sum"
            ).item()  # sum up batch loss
            pred = output.argmax(
                dim=1, keepdim=True
            )  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print(
        "\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n".format(
            test_loss,
            correct,
            len(test_loader.dataset),
            100.0 * correct / len(test_loader.dataset),
        )
    )


def main():
    # Training settings
    parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--test-batch-size",
        type=int,
        default=1000,
        metavar="N",
        help="input batch size for testing (default: 1000)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=5,
        metavar="N",
        help="number of epochs to train (default: 14)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=1.0,
        metavar="LR",
        help="learning rate (default: 1.0)",
    )
    parser.add_argument(
        "--gamma",
        type=float,
        default=0.7,
        metavar="M",
        help="Learning rate step gamma (default: 0.7)",
    )
    parser.add_argument(
        "--no-cuda", action="store_true", default=False, help="disables CUDA training"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="quickly check a single pass",
    )
    parser.add_argument(
        "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
    )
    parser.add_argument(
        "--log-interval",
        type=int,
        default=10,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument(
        "--save-model",
        action="store_true",
        default=False,
        help="For Saving the current Model",
    )
    args = parser.parse_args()
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    train_kwargs = {"batch_size": args.batch_size}
    test_kwargs = {"batch_size": args.test_batch_size}
    if use_cuda:
        cuda_kwargs = {"num_workers": 1, "pin_memory": True, "shuffle": True}
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)

    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    dataset1 = datasets.MNIST("../data", train=True, download=True, transform=transform)
    dataset2 = datasets.MNIST("../data", train=False, transform=transform)
    train_loader = DataLoader(dataset1, **train_kwargs)
    test_loader = DataLoader(dataset2, **test_kwargs)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=args.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=args.gamma)

    train(args, model, device, train_loader, test_loader, optimizer, scheduler)


if __name__ == "__main__":
    main()



"""IO Pickle module."""

import pickle

from perun.data_model.data import DataNode


def exportPickle(dataNode: DataNode) -> bytes:
    """Export data node to pickle file.

    Parameters
    ----------
    dataNode : DataNode
        Data Node

    Returns
    -------
    bytes
        Binary data to write to file.
    """
    return pickle.dumps(dataNode)


def importPickle(pickleData: bytes) -> DataNode:
    """Import DataNode from pickled data file.

    Parameters
    ----------
    pickleData : bytes
        Binary Data

    Returns
    -------
    DataNode
        DataNode
    """
    obj = pickle.loads(pickleData)
    if isinstance(obj, DataNode):
        return obj
    else:
        raise ValueError("Invalid pickle")



"""IO Json module."""

import json

from perun.data_model.data import DataNode
from perun.io.util import NumpyEncoder


def exportJson(dataNode: DataNode) -> str:
    """Export DataNode to json.

    :param dataNode: DataNode
    :type dataNode: DataNode
    :return: Json string of data node.
    :rtype: str
    """
    dataDict = dataNode.toDict(True)
    return json.dumps(dataDict, cls=NumpyEncoder)


def importJson(jsonString: str) -> DataNode:
    """Create DataNode from JSON string."""
    return DataNode.fromDict(json.loads(jsonString))



"""Device module."""

import enum
from dataclasses import asdict, dataclass
from typing import Any, Callable

from perun.data_model.measurement_type import MetricMetaData, Number


class DeviceType(str, enum.Enum):
    """DeviceType enum."""

    RACK = "rack"
    NODE = "node"
    CPU = "cpu"
    GPU = "gpu"
    RAM = "ram"
    DISK = "disk"
    NET = "net"
    FAN = "fan"
    OTHER = "other"


@dataclass
class Sensor:
    """Defines a devices sensor properties."""

    id: str
    type: DeviceType
    metadata: dict
    dataType: MetricMetaData
    measureCallback: Callable[[], Number]

    def read(self) -> Number:
        """Read value from sensor."""
        return self.measureCallback()

    def toDict(self) -> dict[str, Any]:
        """Return device as a dictionary."""
        return {
            "id": self.id,
            "type": self.type.value,
            "metadata": self.metadata,
            "dataType": asdict(self.dataType),
        }



import random
from unittest.mock import MagicMock

from hypothesis import given
from hypothesis import strategies as st

from perun.comm import Comm
from perun.coordination import assignSensors, getHostRankDict


@given(
    hostnames=st.lists(
        st.tuples(st.text(), st.integers(min_value=1, max_value=5)), min_size=1
    ),
    available_sensors=st.dictionaries(st.text(), st.tuples(st.text(), st.text())),
)
def test_assignSensors(hostnames, available_sensors):
    # First test getHostRankDict
    hostnames_list = []
    for hostname, n in hostnames:
        hostnames_list.extend([hostname] * n)

    world_size = len(hostnames_list)
    comm = MagicMock(spec=Comm)
    comm.Get_rank = MagicMock(return_value=random.randint(0, world_size - 1))
    comm.Get_size = MagicMock(return_value=world_size)
    comm.allgather = MagicMock(
        return_value=[(hostnames_list[rank], rank) for rank in range(world_size)]
    )

    hostRankDict_result = getHostRankDict(comm, hostnames_list[comm.Get_rank()])

    assert isinstance(hostRankDict_result, dict)

    assigned_ranks = set()
    for key, value in hostRankDict_result.items():
        assert isinstance(key, str)
        assert isinstance(value, list)
        assert all(isinstance(rank, int) for rank in value)
        # Ensure that the ranks are unique across all hostnames, and are within the world size
        for rank in value:
            assert rank not in assigned_ranks
            assigned_ranks.add(rank)
            assert rank < world_size
            assert rank >= 0

    comm.Get_rank.assert_called()
    comm.allgather.assert_called()

    g_available_sensors = [available_sensors for _ in range(world_size)]

    assignedSensors_result = assignSensors(hostRankDict_result, g_available_sensors)
    assert isinstance(assignedSensors_result, list)
    assert len(assignedSensors_result) == world_size

    # Ensure that only the first rank of each host has sensors assigned
    for hostname, ranks in hostRankDict_result.items():
        for rank in ranks:
            if rank == min(ranks):
                assert assignedSensors_result[rank] == available_sensors
            else:
                assert assignedSensors_result[rank] == {}



from perun.util import filter_sensors, increaseIdCounter


def test_increaseIdCounter_no_existing_ids():
    existing = []
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test"


def test_increaseIdCounter_existing_ids_no_match():
    existing = ["test_1", "test_2"]
    newId = "new_test"
    result = increaseIdCounter(existing, newId)
    assert result == "new_test"


def test_increaseIdCounter_existing_ids_match():
    existing = ["test", "test_1", "test_2"]
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test_3"


def test_increaseIdCounter_mixed_existing_ids():
    existing = ["test", "test_1", "other_test", "test_2"]
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test_3"


def test_increaseIdCounter_no_match_with_suffix():
    existing = ["test_1", "test_2", "test_3"]
    newId = "test_4"
    result = increaseIdCounter(existing, newId)
    assert result == "test_4"


def test_increaseIdCounter_existing_ids_with_suffix():
    existing = ["test_1", "test_2", "test_3"]
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test_4"


def test_increaseIdCounter_existing_ids_with_missing_entries():
    existing = ["test_1", "test", "test_3", "test_10"]
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test_11"

    existing = ["test_10"]
    newId = "test"
    result = increaseIdCounter(existing, newId)
    assert result == "test_11"


def test_increaseIdCounter_double_suffix():
    existing = ["test_1", "test_2", "test_3", "test_3_1", "test_3_2"]
    newId = "test_2"
    result = increaseIdCounter(existing, newId)
    assert result == "test_2_1"

    existing = ["test_1", "test_2", "test_3", "test_3_1", "test_3_2"]
    newId = "test_3"
    result = increaseIdCounter(existing, newId)
    assert result == "test_3_3"


def test_filter_sensors_no_filters():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    result = filter_sensors(sensors)
    assert result == sensors


def test_filter_sensors_include_sensors():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    include_sensors = ["sensor1", "sensor3"]
    result = filter_sensors(sensors, include_sensors=include_sensors)
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}


def test_filter_sensors_exclude_sensors():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    exclude_sensors = ["sensor2"]
    result = filter_sensors(sensors, exclude_sensors=exclude_sensors)
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}


def test_filter_sensors_include_and_exclude_sensors():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    include_sensors = ["sensor1", "sensor2"]
    exclude_sensors = ["sensor2"]
    result = filter_sensors(
        sensors, include_sensors=include_sensors, exclude_sensors=exclude_sensors
    )
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}


def test_filter_sensors_include_backends():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    include_backends = ["backend1"]
    result = filter_sensors(sensors, include_backends=include_backends)
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}


def test_filter_sensors_exclude_backends():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    exclude_backends = ["backend2"]
    result = filter_sensors(sensors, exclude_backends=exclude_backends)
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}


def test_filter_sensors_include_and_exclude_backends():
    sensors = {
        "sensor1": ("backend1",),
        "sensor2": ("backend2",),
        "sensor3": ("backend1",),
    }
    include_backends = ["backend1", "backend2"]
    exclude_backends = ["backend2"]
    result = filter_sensors(
        sensors, include_backends=include_backends, exclude_backends=exclude_backends
    )
    assert result == {"sensor1": ("backend1",), "sensor3": ("backend1",)}



"""Pandas IO module."""

from pathlib import Path
from typing import Any

import pandas as pd

from perun.data_model.data import DataNode


def exportCSV(outputPath: Path, dataNode: DataNode, mr_id: str) -> None:
    """Export data node to csv format.

    Parameters
    ----------
    outputPath : Path
        Path to export data to.
    dataNode : DataNode
        Perun data node.
    mr_id : str
        Id of Multi_run node to get data from
    """
    columns = [
        "run id",
        "hostname",
        "device_group",
        "sensor",
        "unit",
        "magnitude",
        "timestep",
        "value",
    ]

    rows = []
    mrNode = dataNode.nodes[mr_id]
    for run_n, runNode in mrNode.nodes.items():
        rows.extend(_rowsFromRunNode(runNode, run_n))

    df = pd.DataFrame(rows, columns=columns)
    df.to_csv(outputPath)


def _rowsFromRunNode(runNode: DataNode, run_n: int) -> list[tuple[Any, ...]]:
    """Create table rows from a RUN type data node.

    Parameters
    ----------
    runNode : DataNode
        RUN type node
    run_n : int
        Id number of data node

    Returns
    -------
    list[tuple[Any, ...]]
        List of tuples with table entries.
    """
    rows: list[tuple[Any, ...]] = []
    for hostId, hostNode in runNode.nodes.items():
        for deviceGroupId, deviceGroupNode in hostNode.nodes.items():
            for sensorId, sensorNode in deviceGroupNode.nodes.items():
                if sensorNode.raw_data is not None:
                    rawData = sensorNode.raw_data
                    unit = rawData.v_md.unit.value
                    mag = rawData.v_md.mag.value
                    for i in range(len(rawData.timesteps)):
                        rows.append(
                            (
                                run_n,
                                hostId,
                                deviceGroupId,
                                sensorId,
                                unit,
                                mag,
                                rawData.timesteps[i],
                                rawData.values[i],
                            )
                        )

    return rows



"""Coordination module."""

import logging

from perun.comm import Comm

log = logging.getLogger(__name__)


def getHostRankDict(comm: Comm, hostname: str) -> dict[str, list[int]]:
    """Return a dictionary with all the host names with each MPI rank in them.

    Parameters
    ----------
    comm : Comm
        MPI Communicator
    hostname : str
        Local rank Hostname

    Returns
    -------
    dict[str, list[int]]
        Global host and mpi ranks dictionary.
    """
    rank = comm.Get_rank()

    gHostRank: list[tuple[str, int]] = comm.allgather((hostname, rank))
    hostRankDict: dict[str, list[int]] = {}
    for h, r in gHostRank:
        if h in hostRankDict:
            hostRankDict[h].append(r)
        else:
            hostRankDict[h] = [r]

    return hostRankDict


def assignSensors(
    host_rank_dict: dict[str, list[int]],
    g_available_sensors: list[dict[str, tuple]],
) -> list[dict[str, tuple]]:
    """Assings each mpi rank a sensor based on available backends and Host to rank mapping.

    Parameters
    ----------
    host_rank_dict : dict[str, list[int]]
        Host to rank mapping.
    g_available_sensors : list[dict[str, tuple]]
        List of available sensors for each backend for each rank.

    Returns
    -------
    list[dict[str, tuple]]
        List with apointed backend and sensors for each MPI rank.
    """
    g_assigned_sensors: list[dict[str, tuple]] = [
        {} for _ in range(len(g_available_sensors))
    ]
    for _, ranks in host_rank_dict.items():
        firstRank = sorted(ranks)[0]
        merged_sensors: dict[str, tuple] = {}
        for rank in ranks:
            merged_sensors.update(g_available_sensors[rank])

        g_assigned_sensors[firstRank] = merged_sensors
    return g_assigned_sensors



"""Logging utilities."""

import logging
import pathlib
import sys
from configparser import ConfigParser

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

COLORS = {
    "WARNING": YELLOW,
    "INFO": WHITE,
    "DEBUG": GREEN,
    "CRITICAL": RED,
    "ERROR": RED,
}


class ColoredFormatter(logging.Formatter):
    """A logging formatter that adds colors to specific parts of the log message."""

    def format(self, record: logging.LogRecord) -> str:
        """Apply color formating to log message."""
        levelname = record.levelname
        color_level = COLOR_SEQ % (30 + COLORS.get(levelname, 0))
        record.levelname = f"{color_level}{levelname}{RESET_SEQ}"

        # Add colors to specific fields
        record.asctime = (
            f"{COLOR_SEQ % (30 + GREEN)}{self.formatTime(record)}{RESET_SEQ}"
        )
        record.name = f"{COLOR_SEQ % (30 + CYAN)}{record.name}{RESET_SEQ}"
        record.funcName = f"{COLOR_SEQ % (30 + MAGENTA)}{record.funcName}{RESET_SEQ}"
        record.msg = f"{color_level}{record.msg}{RESET_SEQ}"

        return super().format(record)


def set_logger_config(config: ConfigParser) -> None:
    """
    Configure the logging settings for the application.

    Parameters
    ----------
    config: ConfigParser
        Configuration object.

    Returns
    -------
    None
    """
    # Get logging configuration
    log_lvl = config.get("debug", "log_lvl")
    log_file = config.get("debug", "log_file")

    try:
        from mpi4py import MPI

        rank = f"R{MPI.COMM_WORLD.Get_rank()}/{MPI.COMM_WORLD.Get_size()}:"
    except ImportError:
        rank = ""

    base_logger = logging.getLogger("perun")
    format_string = (
        f"[%(asctime)s][%(name)s][%(funcName)s][%(levelname)s] - {rank}%(message)s"
    )

    simple_formatter = logging.Formatter(fmt=format_string)

    # Std out
    if log_file is None:
        std_handler = logging.StreamHandler(stream=sys.stdout)
        formatter = ColoredFormatter(fmt=format_string)
        std_handler.setFormatter(formatter)
        base_logger.addHandler(std_handler)
    else:
        log_path = pathlib.Path(log_file)
        log_dir = log_path.parents[0]
        log_dir.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(filename=log_path)
        file_handler.setFormatter(simple_formatter)
        base_logger.addHandler(file_handler)

    base_logger.setLevel(log_lvl)



"""Unit module."""

import dataclasses
import enum
import logging
from typing import Any

import numpy as np

log = logging.getLogger(__name__)

Number = int | float | np.integer | np.floating


class Unit(str, enum.Enum):
    """Unit enum."""

    JOULE = "J"
    WATT = "W"
    BYTE = "B"
    SECOND = "s"
    PERCENT = "%"
    SCALAR = ""
    GRAM = "g"
    HZ = "Hz"

    @property
    def symbol(self) -> str:
        """Symbol associated with Unit.

        Returns
        -------
        str
            Unit symbol string.
        """
        return self.value

    def __str__(self) -> str:
        """Convert object to string."""
        return self.symbol


_mag_symbols: dict[str, str] = {
    "PICO": "p",
    "NANO": "n",
    "MICRO": "µ",
    "MILI": "m",
    "ONE": "",
    "KILO": "k",
    "MEGA": "M",
    "GIGA": "G",
    "TERA": "T",
}


class Magnitude(float, enum.Enum):
    """Magnitude prefix enum."""

    PICO = 1e-12
    NANO = 1e-9
    MICRO = 1e-6
    MILI = 1e-3
    ONE = 1
    KILO = 1e3
    MEGA = 1e6
    GIGA = 1e9
    TERA = 1e12

    @classmethod
    def fromSymbol(cls, symbol: str) -> "Magnitude":
        """Create a Magniture objet from a magnigure symbol.

        Parameters
        ----------
        symbol : str
            Magnigure symbol string.

        Returns
        -------
        _type_
            Magniture Enum object.

        Raises
        ------
        ValueError
            If an invalid symbol is given.
        """
        for key, value in _mag_symbols.items():
            if value == symbol:
                return cls[key]
        msg = f"Magnitude symbol {symbol} does not exist."
        log.error(msg)
        raise ValueError(msg)

    @property
    def symbol(self) -> str:
        """Symbol associated with magnitude prefix.

        Returns
        -------
        str
            String symbol
        """
        return _mag_symbols[self.name]

    def __str__(self) -> str:
        """Convert object to string."""
        return self.symbol


@dataclasses.dataclass
class MetricMetaData:
    """Collects a metric metadata."""

    unit: Unit
    mag: Magnitude
    dtype: np.dtype
    min: Number
    max: Number
    fill: Number

    @classmethod
    def fromDict(cls, mdDict: dict[str, Any]) -> "MetricMetaData":
        """Create MetricMetadata from a dictionary."""
        dtype = np.dtype(mdDict["dtype"])
        return cls(
            Unit(mdDict["unit"]),
            Magnitude(mdDict["mag"]),
            dtype,
            dtype.type(mdDict["min"]),
            dtype.type(mdDict["max"], dtype=dtype),
            dtype.type(mdDict["fill"], dtype=dtype),
        )

    def copy(self) -> "MetricMetaData":
        """Copy MetricMetaData object.

        Returns
        -------
        _type_
            Copy of object.
        """
        return MetricMetaData(
            Unit(self.unit.value),
            Magnitude(self.mag.value),
            self.dtype,
            self.min,
            self.max,
            self.fill,
        )



"""Processing Module."""

import copy
import logging
from configparser import ConfigParser
from datetime import datetime
from itertools import chain
from typing import Any

import numpy as np
import numpy.typing as npt

from perun.data_model.data import (
    AggregateType,
    DataNode,
    Metric,
    MetricType,
    NodeType,
    RawData,
    Region,
    Stats,
)
from perun.data_model.measurement_type import Magnitude, MetricMetaData, Number, Unit
from perun.data_model.sensor import DeviceType

log = logging.getLogger(__name__)


def processEnergyData(
    raw_data: RawData,
    start: Number | None = None,
    end: Number | None = None,
) -> tuple[Any, Any]:
    """Calculate total energy and average power from an energy or power time series.

    Using the start and end parameters the results can be limited to certain areas of the application run.

    Parameters
    ----------
    raw_data : RawData
        Raw Data from sensor
    start : Optional[Number], optional
        Start time of region, by default None
    end : Optional[Number], optional
        End time of region, by default None

    Returns
    -------
    _type_
       Tuple with total energy in joules and avg power in watts.
    """
    t_s = raw_data.timesteps.astype("float32")
    t_s *= raw_data.t_md.mag.value / Magnitude.ONE.value
    magFactor = raw_data.v_md.mag.value / Magnitude.ONE.value

    if raw_data.v_md.unit == Unit.JOULE:
        # If getting energy, transform to power
        e_J = raw_data.values
        maxValue = raw_data.v_md.max
        dtype = raw_data.v_md.dtype.name

        d_energy = np.diff(e_J)

        if "uint" in dtype:
            idx = d_energy >= maxValue
            max_dtype = np.iinfo(dtype).max
            d_energy[idx] = maxValue + d_energy[idx] - max_dtype
        else:
            idx = d_energy <= 0
            d_energy[idx] = d_energy[idx] + maxValue

        d_energy = d_energy.astype("float32")

        # Transform the energy series to a power series
        power_W = d_energy / np.diff(t_s)
        power_W = np.insert(power_W, 0, power_W[0])
        power_W *= magFactor

        raw_data.alt_values = e_J
        raw_data.alt_v_md = raw_data.v_md

        raw_data.values = power_W
        raw_data.v_md = MetricMetaData(
            Unit.WATT,
            Magnitude.ONE,
            np.dtype("float32"),
            np.float32(0),
            np.finfo("float32").max,
            np.float32(-1),
        )

    elif raw_data.v_md.unit == Unit.WATT:
        power_W = raw_data.values.astype("float32") * magFactor

    if start and end:
        t_s, power_W = getInterpolatedValues(t_s, power_W, start, end)

    avg_power_W = np.mean(power_W)
    if np.__version__[0] == "1":
        energy_J = np.trapz(power_W, x=t_s)  # type: ignore[attr-defined]
    else:
        energy_J = np.trapezoid(power_W, x=t_s)  # type: ignore[attr-defined]

    return energy_J, avg_power_W


def processSensorData(sensorData: DataNode) -> DataNode:
    """Calculate metrics based on raw values.

    Parameters
    ----------
    sensorData : DataNode
        DataNode with raw sensor data.

    Returns
    -------
    DataNode
        DataNode with computed metrics.
    """
    if sensorData.type == NodeType.SENSOR and sensorData.raw_data:
        rawData = sensorData.raw_data

        runtime: float = rawData.timesteps[-1].item()
        sensorData.metrics[MetricType.RUNTIME] = Metric(
            MetricType.RUNTIME, runtime, rawData.t_md, AggregateType.MAX
        )

        if rawData.v_md.unit == Unit.JOULE or rawData.v_md.unit == Unit.WATT:
            energy_J, power_W = processEnergyData(rawData)

            energyMetric = Metric(
                MetricType.ENERGY,
                energy_J,
                MetricMetaData(
                    Unit.JOULE,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )
            powerMetric = Metric(
                MetricType.POWER,
                power_W,
                MetricMetaData(
                    Unit.WATT,
                    Magnitude.ONE,
                    np.dtype("float32"),
                    np.float32(0),
                    np.finfo("float32").max,
                    np.float32(-1),
                ),
                AggregateType.SUM,
            )

            sensorData.metrics[MetricType.ENERGY] = energyMetric
            sensorData.metrics[MetricType.POWER] = powerMetric

            if sensorData.deviceType == DeviceType.CPU:
                sensorData.metrics[MetricType.CPU_ENERGY] = energyMetric.copy()
                sensorData.metrics[MetricType.CPU_ENERGY].type = MetricType.CPU_ENERGY
                sensorData.metrics[MetricType.CPU_POWER] = powerMetric.copy()
                sensorData.metrics[MetricType.CPU_POWER].type = MetricType.CPU_POWER

            elif sensorData.deviceType == DeviceType.GPU:
                sensorData.metrics[MetricType.GPU_ENERGY] = energyMetric.copy()
                sensorData.metrics[MetricType.GPU_ENERGY].type = MetricType.GPU_ENERGY
                sensorData.metrics[MetricType.GPU_POWER] = powerMetric.copy()
                sensorData.metrics[MetricType.GPU_POWER].type = MetricType.GPU_POWER

            elif sensorData.deviceType == DeviceType.RAM:
                sensorData.metrics[MetricType.DRAM_ENERGY] = energyMetric.copy()
                sensorData.metrics[MetricType.DRAM_ENERGY].type = MetricType.DRAM_ENERGY
                sensorData.metrics[MetricType.DRAM_POWER] = powerMetric.copy()
                sensorData.metrics[MetricType.DRAM_POWER].type = MetricType.DRAM_POWER

            elif sensorData.deviceType == DeviceType.OTHER:
                sensorData.metrics[MetricType.OTHER_ENERGY] = energyMetric.copy()
                sensorData.metrics[
                    MetricType.OTHER_ENERGY
                ].type = MetricType.OTHER_ENERGY
                sensorData.metrics[MetricType.OTHER_POWER] = powerMetric.copy()
                sensorData.metrics[MetricType.OTHER_POWER].type = MetricType.OTHER_POWER

        elif rawData.v_md.unit == Unit.PERCENT:
            if sensorData.deviceType == DeviceType.CPU:
                metricType = MetricType.CPU_UTIL
            elif sensorData.deviceType == DeviceType.GPU:
                metricType = MetricType.GPU_UTIL
            else:
                metricType = MetricType.OTHER_UTIL

            sensorData.metrics[metricType] = Metric(
                metricType,
                np.mean(rawData.values),  # type: ignore[arg-type]
                rawData.v_md,
                AggregateType.MEAN,
            )
        elif rawData.v_md.unit == Unit.BYTE:
            bytes_v = rawData.values

            if sensorData.deviceType == DeviceType.NET:
                if "READ" in sensorData.id:
                    metricType = MetricType.NET_READ
                else:
                    metricType = MetricType.NET_WRITE

                d_bytes = bytes_v[1:] - bytes_v[:-1]
                result = d_bytes.sum()
                aggType = AggregateType.SUM
            elif sensorData.deviceType == DeviceType.DISK:
                if "READ" in sensorData.id:
                    metricType = MetricType.DISK_READ
                else:
                    metricType = MetricType.DISK_WRITE

                d_bytes = bytes_v[1:] - bytes_v[:-1]
                result = d_bytes.sum()
                aggType = AggregateType.SUM
            elif sensorData.deviceType == DeviceType.GPU:
                metricType = MetricType.GPU_MEM
                result = bytes_v.mean()
                aggType = AggregateType.SUM
            elif sensorData.deviceType == DeviceType.RAM:
                metricType = MetricType.DRAM_MEM
                result = bytes_v.mean()
                aggType = AggregateType.SUM
            else:
                metricType = MetricType.OTHER_MEM
                result = bytes_v.mean()
                aggType = AggregateType.SUM

            sensorData.metrics[metricType] = Metric(
                metricType, result.astype(rawData.v_md.dtype), rawData.v_md, aggType
            )

        sensorData.processed = True
    return sensorData


def processDataNode(
    dataNode: DataNode, perunConfig: ConfigParser, force_process: bool = False
) -> DataNode:
    """Recursively calculate metrics on the dataNode tree.

    Parameters
    ----------
    dataNode : DataNode
        Root data node tree.
    perunConfig: ConfigParser
        Perun configuration
    force_process : bool, optional
        Force recomputation of child node metrics, by default False

    Returns
    -------
    DataNode
        Data node with computed metrics.
    """
    # Regions
    if dataNode.regions:
        start = datetime.now()
        unprocessedRegions = []
        for region in dataNode.regions.values():
            if not region.processed:
                addRunAndRuntimeInfoToRegion(region)
                region.processed = True
                unprocessedRegions.append(region)

        processRegionsWithSensorData(unprocessedRegions, dataNode)
        duration = datetime.now() - start
        log.info(f"Region processing duration: {duration}")

    aggregatedMetrics: dict[MetricType, list[Metric]] = {}
    for _, subNode in dataNode.nodes.items():
        # Make sure sub nodes have their metrics ready
        if not subNode.processed or force_process:
            if subNode.type == NodeType.SENSOR:
                subNode = processSensorData(subNode)
            else:
                subNode = processDataNode(
                    subNode, perunConfig=perunConfig, force_process=force_process
                )

        if dataNode.type == NodeType.APP:
            for subSubNode in subNode.nodes.values():
                for metricType, metric in subSubNode.metrics.items():
                    if isinstance(metric, Metric):
                        if metricType in aggregatedMetrics:
                            aggregatedMetrics[metricType].append(metric)
                        else:
                            aggregatedMetrics[metricType] = [metric]

        else:
            for metricType, metric in subNode.metrics.items():
                if isinstance(metric, Metric):
                    if metricType in aggregatedMetrics:
                        aggregatedMetrics[metricType].append(metric)
                    else:
                        aggregatedMetrics[metricType] = [metric]

    for metricType, metrics in aggregatedMetrics.items():
        aggType = metrics[0].agg
        metric_md = metrics[0].metric_md
        if dataNode.type == NodeType.MULTI_RUN or dataNode.type == NodeType.APP:
            dataNode.metrics[metricType] = Stats.fromMetrics(metrics)
        else:
            if aggType == AggregateType.MEAN:
                aggregatedValue = np.array([metric.value for metric in metrics]).mean()
            elif aggType == AggregateType.MAX:
                aggregatedValue = np.array([metric.value for metric in metrics]).max()
            elif aggType == AggregateType.MIN:
                aggregatedValue = np.array([metric.value for metric in metrics]).min()
            else:
                aggregatedValue = np.array([metric.value for metric in metrics]).sum()

            dataNode.metrics[metricType] = Metric(
                metricType, aggregatedValue, metric_md, aggType
            )

    # Apply power overhead to each computational node if there is power data available.
    if dataNode.type == NodeType.NODE and MetricType.POWER in dataNode.metrics:
        power_overhead = perunConfig.getfloat("post-processing", "power_overhead")
        dataNode.metrics[MetricType.POWER].value += power_overhead
        runtime = dataNode.metrics[MetricType.RUNTIME].value
        dataNode.metrics[MetricType.ENERGY].value += runtime * power_overhead

    # If there is energy data, apply PUE, and convert to currency and CO2 emmisions.
    if dataNode.type == NodeType.RUN and MetricType.ENERGY in dataNode.metrics:
        pue = perunConfig.getfloat("post-processing", "pue")
        emissions_factor = perunConfig.getfloat("post-processing", "emissions_factor")
        price_factor = perunConfig.getfloat("post-processing", "price_factor")
        total_energy = dataNode.metrics[MetricType.ENERGY].value * pue
        dataNode.metrics[MetricType.ENERGY].value = total_energy
        e_kWh = total_energy / (3600 * 1e3)

        costMetric = Metric(
            MetricType.MONEY,
            e_kWh * price_factor,
            MetricMetaData(
                Unit.SCALAR,
                Magnitude.ONE,
                np.dtype("float32"),
                np.float32(0),
                np.finfo("float32").max,
                np.float32(0),
            ),
            AggregateType.SUM,
        )

        co2Emissions = Metric(
            MetricType.CO2,
            e_kWh * emissions_factor,
            MetricMetaData(
                Unit.GRAM,
                Magnitude.ONE,
                np.dtype("float32"),
                np.float32(0),
                np.finfo("float32").max,
                np.float32(0),
            ),
            AggregateType.SUM,
        )
        dataNode.metrics[MetricType.MONEY] = costMetric
        dataNode.metrics[MetricType.CO2] = co2Emissions

    dataNode.processed = True
    return dataNode


def processRegionsWithSensorData(regions: list[Region], dataNode: DataNode) -> None:
    """Complete region information using sensor data found on the data node (in place op).

    Parameters
    ----------
    regions : List[Region]
        List of regions that use the same data node.
    dataNode : DataNode
        Data node with sensor data.
    """
    log.debug(f"Processing regions with sensor data: {len(regions)}")
    power = [
        {
            rank: [0.0 for _ in range(region.raw_data[rank].shape[0] // 2)]
            for rank in region.raw_data.keys()
        }
        for region in regions
    ]
    cpu_util = copy.deepcopy(power)
    dram_mem = copy.deepcopy(power)
    gpu_mem = copy.deepcopy(power)

    has_gpu = False

    for hostNode in dataNode.nodes.values():
        # Get relevant ranks
        ranks = hostNode.metadata["mpi_ranks"]
        for deviceNode in hostNode.nodes.values():
            if (
                deviceNode.deviceType == DeviceType.CPU
                or deviceNode.deviceType == DeviceType.GPU
                or deviceNode.deviceType == DeviceType.RAM
            ):
                for sensorNode in deviceNode.nodes.values():
                    if sensorNode.raw_data:
                        raw_data = sensorNode.raw_data
                        measuring_unit = raw_data.v_md.unit
                        for region_idx, region in enumerate(regions):
                            for rank in ranks:
                                if rank in region.raw_data:
                                    events = region.raw_data[rank]
                                    for i in range(events.shape[0] // 2):
                                        if (
                                            measuring_unit == Unit.JOULE
                                            or measuring_unit == Unit.WATT
                                        ):
                                            _, power_W = processEnergyData(
                                                raw_data,
                                                events[i * 2].item(),
                                                events[i * 2 + 1].item(),
                                            )
                                            power[region_idx][rank][i] += power_W
                                        elif (
                                            measuring_unit == Unit.PERCENT
                                            and deviceNode.deviceType == DeviceType.CPU
                                        ):
                                            _, values = getInterpolatedValues(
                                                raw_data.timesteps.astype("float32"),
                                                raw_data.values,
                                                events[i * 2].item(),
                                                events[i * 2 + 1].item(),
                                            )
                                            cpu_util[region_idx][rank][i] += np.mean(
                                                values, dtype="float32"
                                            )
                                        elif (
                                            measuring_unit == Unit.BYTE
                                            and deviceNode.deviceType == DeviceType.RAM
                                        ):
                                            _, values = getInterpolatedValues(
                                                raw_data.timesteps.astype("float32"),
                                                raw_data.values,
                                                events[i * 2].item(),
                                                events[i * 2 + 1].item(),
                                            )
                                            dram_mem[region_idx][rank][i] += (
                                                np.mean(values)
                                            ).astype("float32")
                                        elif (
                                            measuring_unit == Unit.BYTE
                                            and deviceNode.deviceType == DeviceType.GPU
                                        ):
                                            has_gpu = True
                                            _, values = getInterpolatedValues(
                                                raw_data.timesteps.astype("float32"),
                                                raw_data.values,
                                                events[i * 2].item(),
                                                events[i * 2 + 1].item(),
                                            )
                                            gpu_mem[region_idx][rank][i] += (
                                                np.mean(values)
                                            ).astype("float32")

    for region_idx, region in enumerate(regions):
        r_power = np.array(list(chain(*power[region_idx].values())))
        r_cpu_util = np.array(list(chain(*cpu_util[region_idx].values())))
        r_gpu_mem = np.array(list(chain(*gpu_mem[region_idx].values())))
        r_dram_mem = np.array(list(chain(*dram_mem[region_idx].values())))

        region.metrics[MetricType.CPU_UTIL] = Stats(
            MetricType.CPU_UTIL,
            MetricMetaData(
                Unit.PERCENT,
                Magnitude.ONE,
                np.dtype("float32"),
                np.float32(0),
                np.float32(100),
                np.float32(-1),
            ),
            r_cpu_util.sum(),
            r_cpu_util.mean(),
            r_cpu_util.std(),
            r_cpu_util.max(),
            r_cpu_util.min(),
        )
        region.metrics[MetricType.DRAM_MEM] = Stats(
            MetricType.DRAM_MEM,
            MetricMetaData(
                Unit.BYTE,
                Magnitude.ONE,
                np.dtype("uint64"),
                np.uint64(0),
                np.uint64(np.iinfo("uint64").max),
                np.uint64(0),
            ),
            r_dram_mem.sum(),
            r_dram_mem.mean(),
            r_dram_mem.std(),
            r_dram_mem.max(),
            r_dram_mem.min(),
        )
        region.metrics[MetricType.POWER] = Stats(
            MetricType.POWER,
            MetricMetaData(
                Unit.WATT,
                Magnitude.ONE,
                np.dtype("float32"),
                np.float32(0),
                np.float32(np.finfo("float32").max),
                np.float32(-1),
            ),
            r_power.sum(),
            r_power.mean(),
            r_power.std(),
            r_power.max(),
            r_power.min(),
        )
        if has_gpu:
            region.metrics[MetricType.GPU_MEM] = Stats(
                MetricType.GPU_MEM,
                MetricMetaData(
                    Unit.BYTE,
                    Magnitude.ONE,
                    np.dtype("uint64"),
                    np.uint64(0),
                    np.uint64(np.iinfo("uint64").max),
                    np.uint64(0),
                ),
                r_gpu_mem.sum(),
                r_gpu_mem.mean(),
                r_gpu_mem.std(),
                r_gpu_mem.max(),
                r_gpu_mem.min(),
            )


def addRunAndRuntimeInfoToRegion(region: Region) -> None:
    """Process run and runtime stats in region objects (in place operation).

    Parameters
    ----------
    region : Region
        Region object
    """
    runs_per_rank = []
    runtime = []

    for rank in region.raw_data.keys():
        events = region.raw_data[rank]
        runs_per_rank.append(events.shape[0] / 2)
        for i in range(1, events.shape[0], 2):
            runtime.append(events[i] - events[i - 1])

    runs_array = np.array(runs_per_rank)
    runtime_array = np.array(runtime)

    region.runs_per_rank = Stats(
        MetricType.N_RUNS,
        MetricMetaData(
            Unit.SCALAR,
            Magnitude.ONE,
            np.dtype("float32"),
            np.float32(0),
            np.finfo("float32").max,
            np.float32(-1),
        ),
        runs_array.sum(),
        runs_array.mean(),
        runs_array.std(),
        runs_array.max(),
        runs_array.min(),
    )

    region.metrics[MetricType.RUNTIME] = Stats(
        MetricType.RUNTIME,
        MetricMetaData(
            Unit.SECOND,
            Magnitude.ONE,
            np.dtype("float32"),
            np.float32(0),
            np.finfo("float32").max,
            np.float32(-1),
        ),
        runtime_array.sum(),
        runtime_array.mean(),
        runtime_array.std(),
        runtime_array.max(),
        runtime_array.min(),
    )


def getInterpolatedValues(
    t: npt.NDArray[np.floating], x: npt.NDArray[np.floating], start: Number, end: Number
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Extract a time range out of a time series, and interpolate the values at the edges.

    Parameters
    ----------
    t : np.ndarray
        Original time steps
    x : np.ndarray
        Original values
    start : Number
        Start of the roi
    end : Number
        End of the roi

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple with the new time steps and values.
    """
    new_t = np.concatenate([[start], t[np.all([t >= start, t <= end], axis=0)], [end]])
    new_x = np.interp(new_t, t, x)
    return new_t, new_x



"""Backend module."""

from typing import Type

from .backend import Backend
from .nvml import NVMLBackend
from .powercap_rapl import PowercapRAPLBackend
from .psutil import PSUTILBackend
from .rocmsmi import ROCMBackend

available_backends: dict[str, Type[Backend]] = {
    "NVMLBackend": NVMLBackend,
    "PowercapRAPLBackend": PowercapRAPLBackend,
    "PSUTILBackend": PSUTILBackend,
    "ROCMBackend": ROCMBackend,
}



"""Decorator module."""

import functools
import logging
from typing import Any, Callable

from perun.configuration import (
    config,
    read_custom_config,
    read_environ,
    sanitize_config,
    save_to_config,
)
from perun.core import Perun
from perun.data_model.data import DataNode
from perun.logging import set_logger_config
from perun.monitoring.application import Application
from perun.processing import Number

log = logging.getLogger(__name__)


def monitor(region_name: str | None = None) -> Callable:
    """Decorate function to monitor its energy usage."""

    def inner_function(func: Callable) -> Callable:
        @functools.wraps(func)
        def func_wrapper(*args: Any, **kwargs: Any) -> Any:
            # Get custom config and kwargs
            region_id = region_name if region_name else func.__name__

            perun = Perun(config)
            if perun.warmup_round:
                func_result = func(*args, **kwargs)
            else:
                log.info(f"Rank {perun.comm.Get_rank()}: Entering '{region_id}'")
                perun.mark_event(region_id)
                func_result = func(*args, **kwargs)
                perun.mark_event(region_id)
                log.info(f"Rank {perun.comm.Get_rank()}: Leaving '{region_id}'")

            return func_result

        return func_wrapper

    return inner_function


def perun(configuration_file: str = "./.perun.ini", **conf_kwargs: Any) -> Callable:
    """Decorate function to monitor its energy usage."""

    def inner_function(func: Callable) -> Callable:
        @functools.wraps(func)
        def func_wrapper(*args: Any, **kwargs: Any) -> Any:
            # 1) Read custom config
            read_custom_config(configuration_file)

            # 2) Read environment variables
            read_environ()

            # 3) Parse remaining arguments
            for key, value in conf_kwargs.items():
                save_to_config(key, value)

            sanitize_config(config)
            set_logger_config(config)

            app = Application(func, config, args=args, kwargs=kwargs)
            perun = Perun(config)

            func_result = perun.monitor_application(app)

            return func_result

        return func_wrapper

    return inner_function


def register_callback(func: Callable[[DataNode], None]) -> None:
    """Register a function to run after perun has finished collection data.

    Parameters
    ----------
    func : Callable[[DataNode], None]
        Function to be called.
    """
    perun: Perun | None = Perun.getInstance()

    if perun and func.__name__ not in perun._postprocess_callbacks:
        log.info(f"Rank {perun.comm.Get_rank()}: Registering callback {func.__name__}")
        perun._postprocess_callbacks[func.__name__] = func


def register_live_callback(
    obj: Callable[[], Callable[[dict[str, Number]], None]],
    id: str,
) -> None:
    """
    Register a function that initializes a live callback function that will be run after each datapoint is collected on the monitoring subprocess.

    This is useful for live monitoring of metrics in real-time.

    The function passed should return a callable that accepts the metric identifier and the metric value.

    This structure is for systems that need to establish a connection to an external server, like MLFlow and Weights and Biases, as such objects are not serializable and can sometimes cause issues with the multiprocessing module.

    Parameters
    ----------
    obj : Callable[[], Callable[[str, Union[int, float]], None]]
        Function that initializes the live callback.
        It should return a callable that accepts the metric identifier and the metric value. It should take no arguments.
    id : str
        Identifier for the live callback, used to register it in the Perun instance.
    """
    perun: Perun | None = Perun.getInstance()
    if perun and id not in perun._live_callbacks:
        log.info(f"Rank {perun.comm.Get_rank()}: Registering live callback {id}")
        perun._live_callbacks[id] = obj



"""Storage Module."""

import dataclasses
import enum
import logging
import time
from dataclasses import asdict
from typing import Any

import numpy as np
import numpy.typing as npt

from perun.data_model.measurement_type import MetricMetaData, Number
from perun.data_model.sensor import DeviceType

log = logging.getLogger(__name__)


class NodeType(enum.Enum):
    """DataNode type enum."""

    APP = enum.auto()
    MULTI_RUN = enum.auto()
    RUN = enum.auto()
    NODE = enum.auto()
    DEVICE_GROUP = enum.auto()
    SENSOR = enum.auto()


class MetricType(str, enum.Enum):
    """Metric Type enum."""

    RUNTIME = "runtime"
    POWER = "power"
    CPU_POWER = "cpu_power"
    GPU_POWER = "gpu_power"
    DRAM_POWER = "dram_power"
    OTHER_POWER = "other_power"
    CPU_UTIL = "cpu_util"
    GPU_UTIL = "gpu_util"
    OTHER_UTIL = "other_util"
    DRAM_MEM = "dram_mem"
    GPU_MEM = "gpu_mem"
    NET_READ = "net_read"
    NET_WRITE = "net_write"
    DISK_READ = "disk_read"
    DISK_WRITE = "disk_write"
    ENERGY = "energy"
    CPU_ENERGY = "cpu_energy"
    GPU_ENERGY = "gpu_energy"
    DRAM_ENERGY = "dram_energy"
    OTHER_ENERGY = "other_energy"
    OTHER_MEM = "other_memory"
    CPU_CLOCK = "cpu_clock"
    GPU_CLOCK = "gpu_clock"
    N_RUNS = "n_runs"
    MONEY = "money"
    CO2 = "co2"

    def __str__(self) -> str:
        """Return string representation of MetricType."""
        return self.value

    def __repr__(self) -> str:
        """Return string representation of MetricType."""
        return self.value

    def fromString(self, value: str) -> "MetricType":
        """Create MetricType from string.

        Parameters
        ----------
        value : str
            MetricType value.

        Returns
        -------
        MetricType
            MetricType object.
        """
        return MetricType(value)


class AggregateType(str, enum.Enum):
    """Types of data aggregation."""

    SUM = "sum"
    MEAN = "mean"
    MAX = "max"
    MIN = "min"


@dataclasses.dataclass
class Metric:
    """Struct with resulting metrics and the metadata."""

    type: MetricType
    value: Number
    metric_md: MetricMetaData
    agg: AggregateType

    @classmethod
    def fromDict(cls, metricDict: dict[str, Any]) -> "Metric":
        """Create RawData object from a dictionary."""
        return cls(
            MetricType(metricDict["type"]),
            np.float32(metricDict["value"]),
            MetricMetaData.fromDict(metricDict["metric_md"]),
            AggregateType(metricDict["agg"]),
        )

    def copy(self) -> "Metric":
        """Create copy metric object.

        Returns
        -------
        _type_
            Copy of object.
        """
        return Metric(
            MetricType(self.type.value),
            self.value,
            self.metric_md.copy(),
            AggregateType(self.agg.value),
        )


@dataclasses.dataclass
class Stats:
    """Collects statistics based on multiple metrics of the same type."""

    type: MetricType
    metric_md: MetricMetaData
    sum: Number
    mean: Number
    std: Number
    max: Number
    min: Number

    @classmethod
    def fromMetrics(cls, metrics: list[Metric]) -> "Stats":
        """Create stats object from list of metrics with the same type.

        Parameters
        ----------
        metrics : list[Metric]
            List of metrics with  the same type.

        Returns
        -------
        Stats
            Stats object.

        Raises
        ------
        Exception
            If metrics are not from the same type.
        """
        type = metrics[0].type
        metric_md = metrics[0].metric_md

        for m in metrics:
            if m.type != type:
                log.error("Metrics given to Stats class do not match")
                raise Exception("Metrics type don't match. Invalid Stats")

        values = np.array([metric.value for metric in metrics])
        sum = values.sum()
        mean = values.mean()
        std = values.std()
        max = values.max()
        min = values.min()
        return cls(type, metric_md, sum, mean, std, max, min)

    @property
    def value(self) -> Number:
        """Value property (mean).

        For compatibility with Metric dataclass.

        Returns
        -------
        Number
            Return the mean value of the stats object.
        """
        return self.mean

    @value.setter
    def value(self, n_value: Number) -> None:
        self.mean = n_value

    @classmethod
    def fromDict(cls, statsDict: dict[str, Any]) -> "Stats":
        """Stats constructor from a dictionory."""
        return cls(
            MetricType(statsDict["type"]),
            MetricMetaData.fromDict(statsDict["metric_md"]),
            statsDict["min"],
            statsDict["mean"],
            statsDict["std"],
            statsDict["max"],
            statsDict["min"],
        )


@dataclasses.dataclass
class RawData:
    """Contains timesteps and recorded values from sensors, including information on the values."""

    timesteps: npt.NDArray[np.floating]
    values: npt.NDArray[np.floating | np.integer]
    t_md: MetricMetaData
    v_md: MetricMetaData
    alt_values: npt.NDArray[np.floating | np.integer] | None = None
    alt_v_md: MetricMetaData | None = None

    @classmethod
    def fromDict(cls, rawDataDict: dict[str, Any]) -> "RawData":
        """Create RawData object from a dictionary.

        Parameters
        ----------
        rawDataDict : Dict
            Dictionary with same keys as RawData object.

        Returns
        -------
        _type_
            RawData object.
        """
        t_md = MetricMetaData.fromDict(rawDataDict["t_md"])
        v_md = MetricMetaData.fromDict(rawDataDict["v_md"])
        alt_v_md = (
            MetricMetaData.fromDict(rawDataDict["alt_v_md"])
            if "alt_v_md" in rawDataDict
            else None
        )
        alt_v_dtype = alt_v_md.dtype if alt_v_md else None
        return cls(
            timesteps=np.array(rawDataDict["timesteps"], dtype=t_md.dtype),
            values=np.array(rawDataDict["values"], dtype=t_md.dtype),
            alt_values=(
                np.array(rawDataDict["alt_values"], dtype=alt_v_dtype)
                if "alt_values" in rawDataDict
                else None
            ),
            t_md=t_md,
            v_md=v_md,
            alt_v_md=alt_v_md,
        )


class LocalRegions:
    """Stores local region data while an application is being monitored."""

    def __init__(self) -> None:
        self._regions: dict[str, list[int]] = {}

    def addEvent(self, region_name: str) -> None:
        """Mark a new event for the named region.

        Parameters
        ----------
        region_name : str
            Region to mark the event from.
        """
        if region_name not in self._regions:
            self._regions[region_name] = []

        self._regions[region_name].append(time.time_ns())

    def isEmpty(self) -> bool:
        """Check if there are any regions marked.

        Returns
        -------
        bool
            True if there are no regions marked.
        """
        return len(self._regions.keys()) == 0

    def __str__(self) -> str:
        """Return string representation of LocalRegions object."""
        return str(self._regions)


@dataclasses.dataclass
class Region:
    """Stores region data from all MPI ranks.

    For each marked region (decorated function), an numpy array with timestamps indicating function starts and ends.
    """

    id: str = ""
    raw_data: dict[int, npt.NDArray[np.floating]] = dataclasses.field(
        default_factory=dict
    )
    runs_per_rank: Stats | None = None
    metrics: dict[MetricType, Stats] = dataclasses.field(default_factory=dict)
    processed: bool = False

    def toDict(self) -> dict[str, Any]:
        """Convert regions to a python dictionary.

        Returns
        -------
        dict[str, dict[int, npt.NDArray[np.floating]]]
            Dictionary with region data.
        """
        result: dict[str, Any] = {
            "id": self.id,
            "raw_data": self.raw_data,
        }

        result["runs_per_rank"] = (
            asdict(self.runs_per_rank) if self.runs_per_rank else None
        )
        result["metrics"] = [asdict(metric) for metric in self.metrics.values()]

        return result

    @classmethod
    def fromDict(cls, regionDictionary: dict[str, Any]) -> "Region":
        """Create Regions object from a dictionary.

        Parameters
        ----------
        regionDictionary: dict[str, dict[int, Any]]
            Region dictionary.

        Returns
        -------
        Regions
            Regions object.
        """
        regionObj = Region()
        regionObj.id = regionDictionary["id"]
        regionObj.raw_data = regionDictionary["raw_data"]
        regionObj.processed = regionDictionary["processed"]
        if regionObj.processed:
            regionObj.metrics = {
                MetricType(metric["type"]): Stats.fromDict(metric)
                for metric in regionDictionary["metrics"]
            }
        return regionObj


class DataNode:
    """Recursive data structure that contains all the information of a monitored application."""

    def __init__(
        self,
        id: str,
        type: NodeType,
        metadata: dict[str, Any] = {},
        nodes: dict[str, Any] | None = None,
        metrics: dict[MetricType, Metric | Stats] | None = None,
        deviceType: DeviceType | None = None,
        raw_data: RawData | None = None,
        regions: dict[str, Region] | None = None,
        processed: bool = False,
    ) -> None:
        """Perun DataNode.

        Parameters
        ----------
        id : str
            Node id.
        type : NodeType
            Node type.
        metadata : Dict
            Node metadata.
        nodes : dict[str, Any] | None, optional
            Children nodes, by default None
        metrics : dict[MetricType, Metric | Stats] | None, optional
            Node metrics, by default None
        deviceType : DeviceType | None, optional
            Node device type, only relevant for leaf nodes, by default None
        raw_data : RawData | None, optional
            Raw data object, only relevant for leaf nodes, by default None
        regions: dict[str, Region], optional
            Dictionary containing region event data, by default None
        processed : bool, optional
            Marks if the node has been processed, by default False
        """
        self.id = id
        self.type = type
        self.metadata: dict[str, Any] = metadata
        self.nodes: dict[str, Any] = nodes if nodes else {}
        self.metrics: dict[MetricType, Metric | Stats] = metrics if metrics else {}
        self.deviceType: DeviceType | None = deviceType
        self.raw_data: RawData | None = raw_data
        self.regions: dict[str, Region] | None = regions
        self.processed = processed

    def addRegionData(self, localRegions: list[LocalRegions], start_time: int) -> None:
        """Add region information to to data node.

        Parameters
        ----------
        localRegions : list[LocalRegions]
            Gathered local regions from all MPI ranks
        start_time : int
            'Official' start time of the run.
        """
        self.regions = {}
        log.debug(f"Local regions: {localRegions}")
        for rank, l_region in enumerate(localRegions):
            if not l_region.isEmpty():
                for region_name, data in l_region._regions.items():
                    if region_name not in self.regions:
                        r = Region()
                        r.id = region_name
                        self.regions[region_name] = r

                    t_s = np.array(data)
                    t_s -= start_time
                    t_s = t_s.astype("float32")
                    t_s *= 1e-9
                    self.regions[region_name].raw_data[rank] = t_s

    def toDict(self, include_raw_data: bool = True) -> dict[str, Any]:
        """Transform object to dictionary."""
        resultsDict = {
            "id": self.id,
            "type": self.type.value,
            "metadata": self.metadata,
            "metrics": {
                type.value: dataclasses.asdict(metric)
                for type, metric in self.metrics.items()
            },
            "regions": (
                {
                    region_name: region.toDict()
                    for region_name, region in self.regions.items()
                }
                if self.regions
                else None
            ),
            "deviceType": self.deviceType,
            "processed": self.processed,
        }
        resultsDict["nodes"] = (
            {key: value.toDict(include_raw_data) for key, value in self.nodes.items()},
        )

        if include_raw_data and self.raw_data:
            resultsDict["raw_data"] = dataclasses.asdict(self.raw_data)

        return resultsDict

    @classmethod
    def fromDict(cls, resultsDict: dict[str, Any]) -> "DataNode":
        """Create dataNode from python dictionary.

        Parameters
        ----------
        resultsDict : Dict
            Dictionary with data node attributes.

        Returns
        -------
        _type_
            DataNode object.
        """
        type = NodeType(resultsDict["type"])
        newResults = cls(
            id=resultsDict["id"],
            type=type,
            metadata=resultsDict["metadata"],
            nodes={
                key: DataNode.fromDict(node)
                for key, node in resultsDict["nodes"].items()
            },
            processed=resultsDict["processed"],
        )
        if "deviceType" in resultsDict:
            newResults.deviceType = DeviceType(resultsDict["deviceType"])

        if "metrics" in resultsDict:
            if type == NodeType.MULTI_RUN:
                newResults.metrics = {
                    MetricType(type): Stats.fromDict(metricDict)
                    for type, metricDict in resultsDict["metrics"].items()
                }
            else:
                newResults.metrics = {
                    MetricType(type): Metric.fromDict(metricDict)
                    for type, metricDict in resultsDict["metrics"].items()
                }
        if "raw_data" in resultsDict:
            newResults.raw_data = RawData.fromDict(resultsDict["raw_data"])

        if "regions" in resultsDict:
            newResults.regions = {
                region_name: Region.fromDict(region_dict)
                for region_name, region_dict in resultsDict["regions"].items()
            }

        return newResults
